#!/usr/bin/env python
# -*- coding: utf-8 -*-
#

import itertools
import logging
import os.path
import time

import h5py
import torch.utils.data

from hakkero.dataset.utils import format_size

logger = logging.getLogger(__name__)

try:
    import msgspec

    json_decode = msgspec.json.decode
    json_encode = msgspec.json.encode
except ModuleNotFoundError:
    import json

    json_decode = json.loads
    json_encode = json.dumps


class IndexedDataset(torch.utils.data.Dataset):
    def __init__(self, path, max_retry=1, retry_sleep=5):
        super().__init__()

        self.path = path
        self.data_path = os.path.join(path, "data.jsonl")
        self.index_path = os.path.join(path, "index.h5")
        self.max_retry = max_retry
        self.retry_sleep = retry_sleep

        self._size = None
        self._length = None

        self.init()

    def init(self):
        with h5py.File(self.index_path, "r") as f:
            self._size = f["index"][-1]
            self._length = len(f["index"]) - 1

    @property
    def size(self):
        return self._size

    def __len__(self):
        return self._length

    def safe_key(self, key):
        if isinstance(key, int):
            if not (-len(self) <= key < len(self)):
                raise IndexError(f"index {key} out of range")
            return key
        elif isinstance(key, slice):
            if key.step is not None:
                raise NotImplementedError("strided indexing not implemented")

            # bound range
            start = min(max(-len(self), 0 if key.start is None else key.start), len(self) - 1)
            stop = min(max(-len(self), len(self) if key.stop is None else key.stop), len(self))

            # remap to positive index
            start = start if start >= 0 else start + len(self)
            stop = stop if stop >= 0 else stop + len(self)
            stop = max(start, stop)

            return slice(start, stop, None)
        else:
            raise TypeError(f"indices must be int or slice, not {type(key)}")

    def __getitem__(self, key):
        key = self.safe_key(key)

        if isinstance(key, slice):
            bytes = self.get_bytes(key, safe=False)
            return self.decode_bytes(key, bytes, safe=False)

        if isinstance(key, int):
            byte = self.get_byte(key, safe=False)
            return self.decode_byte(key, byte, safe=False)

    def read(self, i_or_s, offset, size):
        for retry in itertools.count():
            try:
                # destroy the file identifier to avoid pressure on store
                # buffering=0 to avoid overhead during file.seek() and open()
                with open(self.data_path, "rb", buffering=0) as fin:
                    fin.seek(offset)
                    byte = fin.read(size)

                return byte
            except OSError as e:
                if retry >= self.max_retry:
                    raise OSError(f"reach maximum retry: {retry}, the file system is broken.")

                logger.warning(f"retry loading {self.path}:{i_or_s} in {self.retry_sleep} sec due to error: {e}")
                time.sleep(self.retry_sleep)
            except ValueError as e:
                logger.warning(f"fail to read {self.path}:{i_or_s} due to error: {e}")
                return None

    def get_bounds(self, key):
        for retry in itertools.count():
            try:
                with h5py.File(self.index_path, "r") as fin:
                    return fin["index"][key]
            except OSError as e:
                if retry >= self.max_retry:
                    raise OSError(f"reach maximum retry: {retry}, the file system is broken.")

                logger.warning(f"retry loading {self.index_path} in {self.retry_sleep} sec due to error: {e}")
                time.sleep(self.retry_sleep)
            except ValueError as e:
                logger.warning(f"fail to read {self.index_path} due to error: {e}")
                return None

    def get_byte(self, key, safe=True):
        if safe:
            assert isinstance(key, int)
            key = self.safe_key(key)

        bounds = self.get_bounds(slice(key, key + 2))
        if bounds is None:
            return None

        start, stop = bounds[0], bounds[1]
        return self.read(key, start, stop - start)

    def decode_byte(self, key, byte, safe=True):
        if safe:
            assert isinstance(key, int)
            key = self.safe_key(key)

        if byte is None:
            logger.warning(f"Fail to decode {self.path}:{key}: byte is None")
            return None

        try:
            return json_decode(byte)
        except Exception as e:
            logger.warning(f"Fail to decode {self.path}:{key} due to error {e}, byte: {byte}")
            return None

    def get_bytes(self, key, safe=True):
        if safe:
            assert isinstance(key, slice)
            key = self.safe_key(key)
        if key.start == key.stop:
            # early return empty slice
            return []
        bounds = self.get_bounds(slice(key.start, key.stop + 1))
        if bounds is None:
            return None
        start, stop = bounds[0], bounds[-1]
        byte = self.read(key, start, stop - start)

        if byte is None:
            return [None for _ in bounds[:-1]]
        else:
            shifted_bounds = [b - start for b in bounds]
            return [byte[s:e] for s, e in zip(shifted_bounds[:-1], shifted_bounds[1:])]

    def decode_bytes(self, key, bytes, safe=True):
        if safe:
            assert isinstance(key, slice)
            key = self.safe_key(key)
        if bytes is None:
            logger.warning(f"fail to decode {self.path}:{key}: byte string is None")
            return None
        if not bytes:
            return []

        return [self.decode_byte(i, byte, safe=False) for i, byte in zip(range(key.start, key.stop), bytes)]

    def __repr__(self):
        return f"{self.__class__.__name__}(path={self.path}): {len(self)} entries, {self.pretty_size}"

    @property
    def pretty_size(self):
        return format_size(self.size)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#

import itertools
import math
import random

import numpy as np
from scipy.stats import qmc

# Specifies a target value that is ignored and does not contribute to the input gradient
# see torch.nn.functional.cross_entropy
IGNORE_INDEX = -100


class MultinomialSampler:
    # quasi monte-carlo sampler for better coherence to the expected categorical distribution
    def __init__(self, seed=-1):
        if seed < 0:
            self.sampler = itertools.count()
        else:
            self.sampler = qmc.Sobol(d=1, seed=seed)

    def next(self, weights):
        weight_sum = weights.sum()
        if np.isclose(weight_sum, 0):
            raise StopIteration()

        if isinstance(self.sampler, itertools.count):
            while True:
                i = next(self.sampler) % len(weights)
                if weights[i] > 0:
                    return i
        else:
            p_vals = weights / weight_sum

            residual = 1 - p_vals.sum()
            if residual > 0:
                p_vals[p_vals.argmin()] += residual
            else:
                p_vals[p_vals.argmax()] += residual

            return qmc.MultinomialQMC(p_vals, 1, engine=self.sampler).random(1).argmax()


def random_range(start, stop=None, step=None, seed=0):
    """Generator of non-repeated random permutation with the same interface of python `range`.

    Ref: https://stackoverflow.com/q/53551417

    The random.shuffle(list) and random.sample(list, len(list)) require materialize the lists, which result in a
    long initialization period.
    """
    if stop is None:
        start, stop = 0, start

    if step is None:
        step = 1

    # use a mapping to convert a standard range into the desired range
    mapping = lambda i: (i * step) + start

    # compute the number of numbers in this range
    maximum = int(math.ceil((stop - start) / step))

    # early return with empty range
    if maximum == 0:
        yield from ()
        return

    # seed range with a random integer
    value = random.randint(0, maximum)

    # Construct an offset, multiplier, and modulus for a linear
    # congruential generator. These generators are cyclic and
    # non-repeating when they maintain the properties:
    #
    #   1) "modulus" and "offset" are relatively prime.
    #   2) ["multiplier" - 1] is divisible by all prime factors of "modulus".
    #   3) ["multiplier" - 1] is divisible by 4 if "modulus" is divisible by 4.
    #
    # Pick a random odd-valued offset.
    offset = random.randint(0, maximum) * 2 + 1
    # Pick a multiplier 1 greater than a multiple of 4.
    multiplier = 4 * (maximum // 4) + 1
    # Pick a modulus just big enough to generate all numbers (power of 2).
    modulus = int(2 ** math.ceil(math.log2(maximum)))
    # Track how many random numbers have been returned.
    found = 0
    while found < maximum:
        # If this is a valid value, yield it in generator fashion.
        if value < maximum:
            found += 1
            yield mapping(value)
        # Calculate the next value in the sequence.
        value = (value * multiplier + offset) % modulus


class Range:
    def __init__(self, start, stop, step):
        self.start, self.stop, self.step = start, stop, step

    def __repr__(self):
        return f"Range({self.start}, {self.stop}, {self.step})"

    def iterate(self):
        yield from range(self.start, self.stop, self.step)

    def list(self):
        return list(range(self.start, self.stop, self.step))

    def sub_range(self, split, n_splits):
        # strided split of range
        # e.g., [0, 3, 5, 7, 9] can be split into [0, 5, 9] and [3, 7]
        return Range(self.start + self.step * split, self.stop, self.step * n_splits)

    def random_iterate(self):
        yield from random_range(self.start, self.stop, self.step)

    def __len__(self):
        return math.ceil((self.stop - self.start) / self.step)


class RunningAverage:
    def __init__(self, default=100):
        self._default = default
        self._value = 0
        self._count = 0

    @property
    def value(self):
        if self._count == 0:
            return self._default

        return self._value

    def reset(self):
        self._value = 0
        self._count = 0

    def update(self, x):
        if isinstance(x, (list, tuple)):
            if x:
                self._count += len(x)
                xsum = sum(x, start=0.0)
                self._value += (xsum - len(x) * self._value) / self._count
        else:
            self._count += 1
            self._value += (x - self._value) / self._count

    def __repr__(self):
        return str(self.value)


def format_size(size_bytes):
    if size_bytes == 0:
        return "0B"

    names = ("B", "KB", "MB", "GB", "TB", "PB")

    i = int(math.floor(math.log(size_bytes, 1024)))
    p = math.pow(1024, i)
    s = round(size_bytes / p, 2)
    return f"{s}{names[i]}"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#


LEGACY_KEYS = ["title", "summary", "abstract", "text", "question", "answer", "code"]
FMT_MESSAGES = "[{'role': 'x', 'content': 'x'}, ...]"
FMT_PREFERENCES = "{'context': [{'role': 'x', 'content': 'x'}, ...], 'chosen': 'x', 'rejected': 'x'}"


def check_legacy(data):
    if not isinstance(data, dict):
        return False, "wrong data format, expect {key: value}, " + f"but got {data}"

    if all(s not in data for s in LEGACY_KEYS):
        return False, f"No valid keys in input, expect of: ({LEGACY_KEYS})"

    return True, ""


def check_message(data):
    if not isinstance(data, list) or not all(isinstance(d, dict) for d in data):
        return False, f"messages should be {FMT_MESSAGES}, but got {data}"

    if data[-1]["role"] != "assistant":
        return False, "messages[-1]['role'] should be 'assistant'"

    if data[-2]["role"] != "user":
        return False, "messages[-2]['role'] should be 'user'"

    return True, ""


def check_preference(data):
    if not isinstance(data, dict):
        return False, f"messages should be {FMT_PREFERENCES}, but got {data}"

    if "context" not in data or "chosen" not in data or "rejected" not in data:
        return False, f"messages should be {FMT_PREFERENCES}, but got {data}"

    if not isinstance(data["chosen"], str) or not isinstance(data["rejected"], str):
        return False, "chosen or rejected should be string"

    return check_message(data["context"])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#

import itertools
import random

import torch

from hakkero.dataset.strategy.errors import SegmentationError


def integrous(data, max_length, info, r: random.Random = None):
    """normal way: discard sample that is too long, exceed max_length"""
    # pretrain & sft

    if "pixel_values" in data and "image_grid_thw" in data:
        length = data["input"].nelement()
        assert length == data["label"].nelement()
        if length > max_length:
            raise SegmentationError(f"input length {length} > {max_length}")

        return [
            dict(
                input=data["input"],
                label=data["label"],
                pixel_values=data["pixel_values"],
                image_grid_thw=data["image_grid_thw"],
                used=[info],
            )
        ]

    if "input" in data:
        length = data["input"].nelement()
        assert length == data["label"].nelement()
        if length > max_length:
            raise SegmentationError(f"input length {length} > {max_length}")

        return [dict(input=data["input"], label=data["label"], used=[info])]

    # preference data
    if "inputs" in data:
        for key in data["inputs"].keys():
            length = data["inputs"][key].nelement()
            assert length == data["labels"][key].nelement()
            if length > max_length:
                raise SegmentationError(f"length {length} of `{key} > {max_length}")

        return [dict(inputs=data["inputs"], labels=data["labels"], used=[info])]

    raise ValueError("not supported data: {data}")


def concat(data, max_length, info, r: random.Random = None, prev=None):
    """split long and concat it with previous segment"""
    if "input" not in data and "inputs" in data:
        raise ValueError("preference data only support integrous segmentation")

    if prev is not None:
        # concat input and label with previous segment
        input = torch.cat([prev["input"], prev["label"][-1:], data["input"]])
        label = torch.cat([prev["label"], data["input"][0:1], data["label"]])
        first_used = prev["used"] + [info]
        later_used = [info]
    else:
        input = data["input"]
        label = data["label"]

        first_used = later_used = [info]

    length = input.nelement()
    if length > max_length:
        segments = [
            dict(
                input=input[st : st + max_length],
                label=label[st : st + max_length],
                used=first_used if st == 0 else later_used,
            )
            for st in range(0, length, max_length)
        ]
    else:
        segments = [dict(input=input, label=label, used=first_used)]

    if segments[-1]["input"].nelement() == max_length:
        prev = None
    else:
        # concat the last segment to the next sample
        prev = segments.pop()

    if segments and r is not None:
        r.shuffle(segments)

    return segments, prev


def naive(data, max_length, info, r: random.Random = None):
    if "input" not in data and "inputs" in data:
        raise ValueError("preference data only support integrous segmentation")

    used = [info]

    length = data["input"].nelement()
    if length <= max_length:
        return [dict(input=data["input"], label=data["label"], used=used)]

    segments = [
        dict(input=data["input"][st:ed], label=data["label"][st:ed], used=used)
        for st, ed in map(lambda i: (i, min(i + max_length, length)), range(0, length, max_length))
    ]

    if r is not None:
        r.shuffle(segments)

    return segments


def unbiased(data, max_length, info, r: random.Random = None):
    if "input" not in data and "inputs" in data:
        raise ValueError("preference data only support integrous segmentation")

    assert r is not None

    used = [info]

    length = data["input"].nelement()
    if length <= max_length:
        return [dict(input=data["input"], label=data["label"], used=used)]

    segment_lens = [max_length] * (length // max_length)
    remain = length % max_length
    if remain != 0:
        shift = r.randint(0, max_length - remain)
        segment_lens.append(segment_lens.pop() - shift)
        segment_lens.append(remain + shift)
        # ensure the short segments can appear at any position of the long sequence
        r.shuffle(segment_lens)

    bounds = [0] + list(itertools.accumulate(segment_lens))

    segments = [
        dict(input=data["input"][st:ed], label=data["label"][st:ed], used=used)
        for st, ed in zip(bounds[:-1], bounds[1:])
    ]

    r.shuffle(segments)

    return segments



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#

from os import path

import setuptools
from setuptools import find_packages
from setuptools import setup
from hakkero import __version__

version = int(setuptools.__version__.split(".")[0])
assert version > 30, "requires setuptools > 30"

this_directory = path.abspath(path.dirname(__file__))


def fetch_requirements(filename):
    with open(filename, "r") as fd:
        return [r.strip() for r in fd.readlines()]


def fetch_readme():
    with open("README.md", encoding="utf-8") as fin:
        return fin.read()


setup(
    name="hakkero-dataloader",
    url="https://github.com/ericxsun/hakkero-dataloader",
    keywords="Pytorch LM dataloader",
    version=__version__,
    long_description=fetch_readme(),
    long_description_content_type="text/markdown",
    packages=find_packages(exclude=["examples", "tests"]),
    zip_safe=False,
    install_requires=fetch_requirements("requirements.txt"),
    # https://packaging.python.org/guides/distributing-packages-using-setuptools/#universal-wheels
    options={"bdist_wheel": {"universal": "1"}},
    entry_points={
        "console_scripts": ["hakkero=hakkero.dataset.indexify:main"],
    },
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#

import argparse
import json
import multiprocessing as mp
import os
import shutil
from functools import partial

import h5py

from hakkero import __version__
from hakkero.dataset.logger import logger
from hakkero.dataset.strategy.check import check_legacy
from hakkero.dataset.strategy.check import check_message
from hakkero.dataset.strategy.check import check_preference


def _build_chunk_offsets(filename, start, end, worker_id, n_workers):
    offset = start
    bounds = []
    with open(filename, "rb") as fin:
        fin.seek(offset)

        if start > 0:
            fin.readline()
            offset = fin.tell()

        bounds.append(offset)

        while offset < end:
            logger.info(f"worker: {worker_id}/{n_workers}, offset: {offset}, range in offset [{start}, {end}]")
            line = fin.readline()
            offset += len(line)
            bounds.append(offset)

            try:
                js = json.loads(line)
                _ = js["uid"]
                data = js["data"]

                valid_legacy, _ = check_legacy(data)
                valid_message, _ = check_message(data)
                valid_preference, _ = check_preference(data)

                if valid_legacy or valid_message or valid_preference:
                    continue
                else:
                    logger.error(f"invalid format line: {line}")
                    raise ValueError("invalid format")
            except Exception as e:
                logger.error(f"invalid format line: {line}\n{e}")
                raise e

    return bounds


def build_index(filename, output=None, num_workers=None):
    if num_workers is None:
        num_workers = mp.cpu_count()

    logger.info(f"build indexed with {num_workers} workers for dataset from {filename}")

    file_size = os.path.getsize(filename)
    chunk_size = file_size // num_workers

    pool = mp.Pool(processes=num_workers)

    chunks = [(i * chunk_size, (i + 1) * chunk_size, i + 1, num_workers) for i in range(num_workers)]
    chunks[-1] = (chunks[-1][0], file_size, chunks[-1][2], num_workers)

    func = partial(_build_chunk_offsets, filename)
    results = pool.starmap(func, chunks)

    pool.close()
    pool.join()

    bounds = [0]
    for chunk_offsets in results:
        if bounds[-1] != chunk_offsets[0]:
            bounds.append(chunk_offsets[0])
        bounds.extend(chunk_offsets[1:])

    bounds = sorted(bounds)
    if bounds[-1] != file_size:
        bounds.append(file_size)

    if output is not None:
        os.makedirs(output, exist_ok=True)
    else:
        output = os.path.dirname(filename)

    logger.info(f"build index.h5 into {output}")
    with h5py.File(os.path.join(output, "index.h5"), "w") as hf:
        hf.create_dataset("index", data=bounds)

    logger.info(f"build data.jsonl into {output}")
    if not os.path.exists(os.path.join(output, "data.jsonl")):
        shutil.copyfile(filename, os.path.join(output, "data.jsonl"))


def main():
    parser = argparse.ArgumentParser(description="build index for dataset")
    parser.add_argument("--version", action="version", version=f"%(prog)s {__version__}")
    parser.add_argument("--filename", type=str, help="full filename of jsonl file")
    parser.add_argument("--output", type=str, help="output path for saving data.jsonl and index.h5")
    parser.add_argument("--num_workers", type=int, default=None, help="number of workers")

    args = parser.parse_args()

    build_index(args.filename, args.output, args.num_workers)


if __name__ == "__main__":
    main()
